from pathlib import Path
import psycopg2
from psycopg2 import sql, errors
import random
from itertools import islice
from psycopg2.extras import execute_values
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Iterable
import time
import logging
import json
from functools import wraps

from ankipan import Reader

from ankipan_db import DBManager, db_config, PROJECT_ROOT

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

def retry_on_deadlock(max_attempts=5, initial_backoff=0.1, max_backoff=20.0):
    """
    Retry the decorated function if it raises psycopg2.errors.DeadlockDetected.
    """
    def decorator(fn):
        @wraps(fn)
        def wrapped(*args, **kwargs):
            backoff = initial_backoff
            for attempt in range(1, max_attempts + 1):
                try:
                    return fn(*args, **kwargs)
                except errors.DeadlockDetected:
                    if attempt == max_attempts:
                        logger.error("Deadlock persisted after %d attempts", max_attempts)
                        raise
                    sleep = backoff + random.random() * backoff
                    logger.warning("Deadlock detected; retrying in %.2fs (attempt %d/%d)",
                                   sleep, attempt, max_attempts)
                    time.sleep(sleep)
                    backoff = min(backoff * 2, max_backoff)
        return wrapped
    return decorator


class Parser:
    def __init__(self, lang, db = None):
        self.lang = lang
        if not db:
            self.db = DBManager(lang)
        else:
            self.db = db
        self.reader = Reader(lang)

    def add_source_category(self, source_category_name, metadata = None):
        sql = """
        WITH existing AS (
        SELECT id
        FROM sources
        WHERE parent_id IS NULL AND name = %s
        ),
        ins AS (
        INSERT INTO sources (parent_id, name, metadata, nesting_level, is_leaf)
        SELECT NULL, %s, %s, 0, %s
        WHERE NOT EXISTS (SELECT 1 FROM existing)
        RETURNING id
        )
        SELECT id FROM ins
        UNION ALL
        SELECT id FROM existing;
        """
        conn = self.db.get_safe_conn()
        if metadata is None:
            metadata = {}
        try:
            with conn, conn.cursor() as cur:
                cur.execute(sql, (source_category_name, source_category_name, json.dumps(metadata), False))
                source_category_id = cur.fetchone()[0]
            return source_category_id
        finally:
            self.db.db_pool.putconn(conn)


    def add_source(self,
                   path,
                   source_category_name: str,
                   overwrite=False,
                   source_root_name=None,
                   n_threads=5,
                   index_separators: Iterable[str] = None,
                   replace_chars: Iterable[str] = None,
                   chunk_size: int = 1000,
                   file_match_pattern = None,
                   dir_match_pattern = None,
                   assert_coherence: bool = False,
                   db_workers: int = 3):

        logger.info(f'Adding source {path}')
        path = Path(path)
        if not path.exists():
            raise RuntimeError(f'Path {path} does not exist.')

        setup_conn = self.db.get_safe_conn()
        try:
            with setup_conn:
                with setup_conn.cursor() as cur:
                    cur.execute(
                        "SELECT id FROM sources WHERE name = %s AND nesting_level = 0",
                        (source_category_name,),
                    )
                    row = cur.fetchone()
                    if not row:
                        raise RuntimeError(
                            f'Source Category "{source_category_name}" not defined in db'
                        )
                    source_category_id = row[0]

                    if source_root_name is None:
                        source_root_name = Path(path).stem

                    cur.execute(
                        "SELECT name FROM sources WHERE parent_id = %s", (source_category_id,)
                    )
                    existing_root_sources = [name for (name,) in cur.fetchall()]

                    if path.is_dir():
                        if source_root_name not in existing_root_sources:
                            cur.execute(
                                """
                                INSERT INTO sources (parent_id, name, metadata, nesting_level, is_leaf)
                                VALUES (%s, %s, %s, %s, %s)
                                RETURNING id
                                """,
                                (source_category_id, source_root_name, None, 1, False),
                            )
                        else:
                            cur.execute(
                                "SELECT id FROM sources WHERE name = %s AND parent_id = %s",
                                (source_root_name, source_category_id),
                            )
                        source_root_id = cur.fetchone()[0]
                        is_leaf = False
                    else:
                        source_root_id = source_category_id
                        is_leaf = True

                    if overwrite:
                        self.db.delete_source(cur, source_root_id)
                        cur.execute(
                            """
                            INSERT INTO sources (parent_id, name, metadata, nesting_level, is_leaf)
                            VALUES (%s, %s, %s, %s, %s)
                            ON CONFLICT (parent_id, name) DO UPDATE SET name = EXCLUDED.name
                            RETURNING id
                            """,
                            (source_category_id, source_root_name, None, 1, is_leaf),
                        )
                        source_root_id = cur.fetchone()[0]
                    source_tree = self.db.get_source_tree_for_id(cur, source_root_id)
        finally:
            self.db.db_pool.putconn(setup_conn)

        imported_files = set()
        parent_ids = {parent_id for parent_id, _, _ in source_tree.values()}
        for node_id, (parent_id, name, is_leaf) in source_tree.items():
            if node_id not in parent_ids and is_leaf:
                parts = []
                current_id = node_id
                while current_id != source_category_id and current_id in source_tree:
                    parent_id, nm, _ = source_tree[current_id]
                    parts.append(nm)
                    current_id = parent_id
                imported_files.add(Path("/".join(reversed(parts))))

        file_paths = list(
            self.reader.collect_file_paths(
                path,
                file_match_pattern=file_match_pattern,
                exclude_paths=imported_files,
                dir_match_pattern=dir_match_pattern,
            )
        )

        if len(file_paths) == 0:
            logger.info(f"Source already fully imported")
            return
        logger.info(f"Files to import: {len(file_paths)} (already in DB: {len(imported_files)})")
        if not file_paths:
            logger.info(f'Source "{source_root_name}" filtered out or already fully imported.')
            return

        file_paths_chunks = [
            file_paths[i : i + chunk_size] for i in range(0, len(file_paths), chunk_size)
        ]
        logger.info(f"Processing {len(file_paths_chunks)} chunks")
        disk_root = Path(path).resolve()

        max_outstanding = max(db_workers * 2, 4)

        @retry_on_deadlock(max_attempts=5)
        def _db_task(files_batch):
            conn = self.db.get_safe_conn()
            try:
                with conn.cursor() as c:
                    c.execute(sql.SQL("SET LOCAL search_path TO {}, public")
                              .format(sql.Identifier(self.lang)))
                self.add_files_to_db(files_batch, source_root_id, disk_root, conn)
            finally:
                self.db.db_pool.putconn(conn)

        with ThreadPoolExecutor(max_workers=db_workers, thread_name_prefix="db-writer") as writer:
            futures = []
            for chunk in file_paths_chunks:
                files = self.reader.open_files(
                    chunk,
                    index_separators=index_separators,
                    replace_chars=replace_chars,
                    source_name=source_root_name,
                    assert_coherence=assert_coherence,
                )
                self.reader.process_files(files, save_pos_data=True, n_threads=n_threads)
                futures.append(writer.submit(_db_task, files))

                # mild backpressure
                if len(futures) >= max_outstanding:
                    futures.pop(0).result()

            # collect & re-raise any exceptions
            for f in as_completed(futures):
                f.result()

        logger.info(f'Finished adding source "{source_root_name}"')

    def add_files_to_db(self, files, source_root_id, disk_root, conn):
        schema = self.lang
        logger.info(f"add_files_to_db: Adding {len(files)} files to db")

        def chunked(it, size):
            it = iter(it)
            while True:
                batch = list(islice(it, size))
                if not batch:
                    break
                yield batch

        disk_root = Path(disk_root).resolve()
        cache: dict[tuple[int,str],int] = {}  # dir node cache

        def ensure_dir(cur, parent_id: int, name: str, level: int) -> int:
            key = (parent_id, name)
            if key in cache:
                return cache[key]
            cur.execute(
                sql.SQL("""
                INSERT INTO {}.sources (parent_id, name, metadata, lemma_counts, nesting_level, is_leaf)
                VALUES (%s, %s, %s, %s, %s, %s)
                ON CONFLICT (parent_id, name) DO NOTHING
                RETURNING id""").format(sql.Identifier(schema)),
                (parent_id, name, json.dumps({}), json.dumps({}), level, False)
            )
            row = cur.fetchone()
            if row:
                new_id = row[0]
            else:
                cur.execute(
                    sql.SQL("SELECT id FROM {}.sources WHERE parent_id = %s AND name = %s")
                    .format(sql.Identifier(schema)),
                    (parent_id, name))
                new_id = cur.fetchone()[0]
            cache[key] = new_id
            return new_id

        try:
            with conn.cursor() as cur:
                cur.execute(
                    sql.SQL("SET LOCAL search_path TO {}, public")
                    .format(sql.Identifier(schema))
                )

                unique_lemmas = set()
                word_rows = {}
                all_text_segments = []
                words_in_text_segments = []
                for file in files:
                    path = Path(file.path).resolve()
                    parent = source_root_id

                    base = disk_root if disk_root.is_dir() else disk_root.parent
                    try:
                        rel = path.parent.relative_to(base).parts
                    except ValueError:
                        rel = ()

                    for lvl, part in enumerate(rel, start=2):
                        parent = ensure_dir(cur, parent, part, lvl)

                    leaf_level = len(rel) + (2 if disk_root.is_dir() else 1)
                    leaf_name = path.name
                    cur.execute(
                        sql.SQL("""
                        INSERT INTO {}.sources (parent_id, name, metadata, lemma_counts, nesting_level, is_leaf)
                        VALUES (%s, %s, %s, %s, %s, %s)
                        ON CONFLICT (parent_id, name) DO NOTHING
                        RETURNING id
                        """).format(sql.Identifier(schema)), (parent, leaf_name, json.dumps({}), json.dumps(dict(file.lemma_counts)), leaf_level, True)
                    )
                    row = cur.fetchone()
                    leaf_id = row[0] if row else None
                    if not leaf_id:
                        cur.execute(
                            sql.SQL("SELECT id FROM {}.sources WHERE parent_id = %s AND name = %s")
                            .format(sql.Identifier(schema)),
                            (parent, leaf_name)
                        )
                        leaf_id = cur.fetchone()[0]
                    for i, tokens in enumerate(file.text_segment_components):
                        start_s, end_s = (file.sub_timestamps[i] if file.sub_timestamps else (None, None))
                        all_text_segments.append((start_s, end_s, file.stanza_segments[i].text, i, leaf_id))
                        for tok in tokens:
                            w, p, xp, l = tok["word"], tok["pos"], tok["xpos"], tok["lemma"]
                            unique_lemmas.add(l)
                            word_rows[(w, p)] = (xp, l)
                            words_in_text_segments.append(((w, p), (leaf_id, i)))

                lemma_id_by_lemma = {}
                if unique_lemmas:
                    vals = sorted(unique_lemmas)  # stable order helps
                    for batch in chunked(vals, 5000):
                        execute_values(
                            cur,
                            sql.SQL("""
                                INSERT INTO {}.lemmas (lemma)
                                VALUES %s
                                ON CONFLICT (lemma) DO NOTHING
                            """).format(sql.Identifier(schema)),
                            [(l,) for l in batch],
                            fetch=False,
                        )
                    for batch in chunked(vals, 5000):
                        cur.execute(
                            sql.SQL("SELECT lemma, id FROM {}.lemmas WHERE lemma = ANY(%s)")
                            .format(sql.Identifier(schema)),
                            (batch,),
                        )
                        lemma_id_by_lemma.update(cur.fetchall())

                word_id_by_key = {}
                if word_rows:
                    word_vals = [(w, p, xp, lemma_id_by_lemma[l]) for (w, p), (xp, l) in word_rows.items()]
                    for batch in chunked(word_vals, 5000):
                        execute_values(
                            cur,
                            sql.SQL("""
                                INSERT INTO {}.words (word, pos, xpos, lemma_id)
                                VALUES %s
                                ON CONFLICT (word, pos) DO NOTHING
                            """).format(sql.Identifier(schema)),
                            batch,
                            fetch=False,
                        )

                    for batch in chunked(word_rows.keys(), 5000):
                        rows = execute_values(
                            cur,
                            sql.SQL("""
                                SELECT w.id, v.word, v.pos
                                FROM (VALUES %s) AS v(word, pos)
                                JOIN {}.words w USING (word, pos)
                            """).format(sql.Identifier(schema)),
                            batch,
                            fetch=True,
                        )
                        for wid, w, p in rows:
                            word_id_by_key[(w, p)] = wid

                segment_id_by_key = {}
                if all_text_segments:
                    for batch in chunked(all_text_segments, 2000):
                        rows = execute_values(
                            cur,
                            sql.SQL("""
                                INSERT INTO {}.text_segments (start_s, end_s, text, "index", source_id)
                                VALUES %s
                                ON CONFLICT (source_id, "index")
                                DO UPDATE SET
                                text   = EXCLUDED.text,
                                start_s = COALESCE(EXCLUDED.start_s, {}.text_segments.start_s),
                                end_s   = COALESCE(EXCLUDED.end_s,   {}.text_segments.end_s)
                                RETURNING id, source_id, "index"
                            """).format(sql.Identifier(schema), sql.Identifier(schema), sql.Identifier(schema)),
                            batch,
                            fetch=True,
                        )
                        for tsid, sid, idx in rows:
                            segment_id_by_key[(sid, idx)] = tsid

                if words_in_text_segments:
                    tuples = []
                    for (w, p), (sid, idx) in words_in_text_segments:
                        wid  = word_id_by_key.get((w, p))
                        tsid = segment_id_by_key.get((sid, idx))
                        if wid and tsid:
                            tuples.append((wid, tsid))

                    for batch in chunked(tuples, 10000):
                        execute_values(
                            cur,
                            sql.SQL("""
                                INSERT INTO {}.words_in_text_segments (word_id, text_segment_id)
                                VALUES %s
                                ON CONFLICT (word_id, text_segment_id) DO NOTHING
                            """).format(sql.Identifier(schema)), batch, fetch=False)
            conn.commit()
            logger.info("Finished adding entries to db")

        except Exception:
            conn.rollback()
            logger.exception("DB batch failed; rolled back.")
            raise

    def remove_source(self, source_root_name, conn_ = None):
        if not conn_:
            conn = self.db.get_safe_conn()
        else:
            conn = conn_
        with conn.cursor() as cur:
            cur.execute("SELECT id FROM sources WHERE name = %s", (source_root_name,))
            source_root_id_raw = cur.fetchone()
            if source_root_id_raw:
                self.db.delete_source(cur, source_root_id_raw[0])
            else:
                logger.error(f'No source with the name {source_root_name} in db, not deleting anything.')
        conn.commit()

    def sync_dir(self,
                 sources_path: str,
                 source_category: str,
                 overwrite = False,
                 n_threads = 5,
                 db_workers = 5,
                 index_separators: Iterable[str] = None,
                 replace_chars=None,
                 start_at: str = None,
                 assert_coherence: bool = False,
                 chunk_size = 5000,
                 file_match_pattern = None,
                 dir_match_pattern = None):
        """
        Parse lemmas from path with text data using the ankimelon.Reader.collect_file_paths() method

        Parameters
        ----------
        sources_path: '/' separated Path to a file or directory with textfiles
        source_category: name of subcategory
        overwrite (optional): overwrite existing data
        n_threads (optional): number of threads to parse with
        index_separators (optional): custom separates for source (see help(ankimelon.Reader.open_files))
        replace_chars (optional): List of characters to be removed when processing
        start_at (optional): If parsing was previously interrupted, specify name of file to start with

        """
        starting = False
        paths = [i for i in Path(sources_path).iterdir()]
        if overwrite:
            logger.info(f'Deleting source {source_category}')
            self.remove_source(source_category)
        for i, path in enumerate(paths):
            if start_at:
                if start_at == path.stem:
                    starting = True
                if not starting:
                    logger.info(f'Skipping source "{path.stem}" because start_at "{start_at}" hat not yet been reached')
                    continue
            logger.info(f'Adding source {i}/{len(paths)}')
            while True:
                try:
                    logger.info(f"Adding {path} to {source_category}")
                    self.add_source(path,
                                    source_category,
                                    n_threads = n_threads,
                                    db_workers=db_workers,
                                    index_separators=index_separators,
                                    replace_chars=replace_chars,
                                    chunk_size=chunk_size,
                                    assert_coherence=assert_coherence,
                                    file_match_pattern = file_match_pattern,
                                    dir_match_pattern = dir_match_pattern)
                    break
                except (psycopg2.OperationalError, psycopg2.InterfaceError) as e:
                    logger.error(f"sync fail: {e}")

                    logger.error(f"Operational error for {path}:\n\n{e}")
                    time.sleep(300)
                    self.db = DBManager(self.lang)

        setup_conn = self.db.get_safe_conn()
        try:
            with setup_conn:
                with setup_conn.cursor() as cur:
                    logger.info(f'Aggregating lemma frequencies for {source_category}')
                    cur.execute(
                        "SELECT id FROM sources WHERE name = %s AND nesting_level = 0",
                        (source_category,))
                    source_category_id = cur.fetchone()[0]
                    cur.execute(
                        """
                        WITH RECURSIVE down AS (
                            SELECT id, parent_id, is_leaf, lemma_counts
                            FROM sources
                            WHERE id = %s
                        UNION ALL
                            SELECT s.id, s.parent_id, s.is_leaf, s.lemma_counts
                            FROM sources s
                            JOIN down d ON s.parent_id = d.id
                        ),
                        leaves AS (
                            SELECT id
                            FROM down
                            WHERE is_leaf
                        ),
                        pairs AS (
                            SELECT l.id AS leaf_id, l.id AS ancestor_id
                            FROM leaves l
                        UNION ALL
                            SELECT p.leaf_id, s.parent_id AS ancestor_id
                            FROM pairs p
                            JOIN sources s ON s.id = p.ancestor_id
                            JOIN down    d ON d.id = s.parent_id
                        ),
                        counts AS (
                            SELECT
                                p.ancestor_id AS id,
                                j.key,
                                SUM( (j.value)::numeric )::bigint AS total
                            FROM pairs p
                            JOIN down leaf ON leaf.id = p.leaf_id
                            JOIN LATERAL jsonb_each(COALESCE(leaf.lemma_counts, '{}'::jsonb)) AS j(key, value) ON TRUE
                            GROUP BY p.ancestor_id, j.key
                        ),
                        ranked AS (
                            SELECT
                                id, key, total,
                                ROW_NUMBER() OVER (PARTITION BY id ORDER BY total DESC, key ASC) AS rn
                            FROM counts
                        ),
                        top_counts AS (
                            SELECT id, key, total
                            FROM ranked
                            WHERE rn <= 10000
                        ),
                        agg AS (
                            SELECT
                                id,
                                jsonb_object_agg(key, to_jsonb(total)) AS lemma_counts
                            FROM top_counts
                            GROUP BY id
                        )
                        UPDATE sources s
                        SET lemma_counts = a.lemma_counts
                        FROM agg a
                        JOIN down d ON d.id = a.id
                        WHERE s.id = a.id
                        AND d.is_leaf = FALSE
                        RETURNING s.id, a.lemma_counts;
                        """,
                        (source_category_id,),
                    )

        finally:
            self.db.db_pool.putconn(setup_conn)
        logger.info(f'Finish sync_dir for {source_category}')



from pathlib import Path
import os
from dotenv import load_dotenv, find_dotenv

PROJECT_ROOT = Path(__file__).parent.parent
SOURCES_DIR = PROJECT_ROOT / 'sources'
SCRAPING_DIR = PROJECT_ROOT / 'scraping'

load_dotenv(PROJECT_ROOT / '.env')

db_config = {
    "user": os.getenv("DATABASE_USERNAME"),
    "password": os.getenv("DATABASE_PASSWORD"),
    "host": os.getenv("DATABASE_IP"),
    "port": os.getenv("DATABASE_PORT"),
    "database": os.getenv("DATABASE_NAME")
}

GEMINI_API_KEY = os.getenv("GEMINI_API_KEY")

from .db_manager import DBManager
from .source_parser import Parser



from ankipan_db import DBManager

db = DBManager('jp')
a = db.get_source_list('wikipedia')
print(a)
# import ankipan_db
# ankipan_db.db_config['database'] = 'ankipan_test_db'
# from ankipan_db import DBManager


# db = DBManager('en')

# print("o")
# a = db.get_segments_for_lemmas([],['like'],'source_category_1', 'de')
# print("a")
# print(a)


from psycopg2 import pool, sql
import psycopg2
import hashlib
import random
from collections import Counter
import logging
from functools import partial, wraps
from collections import defaultdict
import time
import inspect
import heapq
from operator import itemgetter

from typing import Dict, Optional, Iterable, Dict, List

import ankipan_db

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

random.seed(42)

N_CONNS = 20

def with_pool_cursor(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        conn = self.get_safe_conn()
        try:
            with conn.cursor() as cur:
                return func(self, cur, *args, **kwargs)
        finally:
            self.db_pool.putconn(conn)
    return wrapper

def with_conditional_pool_conn(func):
    sig = inspect.signature(func)
    has_conn = 'conn' in sig.parameters

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if not has_conn:
            return func(self, *args, **kwargs)

        try:
            bound = sig.bind_partial(self, *args, **kwargs)
        except TypeError:
            return func(self, *args, **kwargs)

        already_has_conn = ('conn' in bound.arguments) and (bound.arguments['conn'] is not None)
        if already_has_conn:
            return func(self, *args, **kwargs)

        conn = self.get_safe_conn()
        try:
            call_kwargs = {k: v for k, v in bound.arguments.items() if k != 'self'}
            call_kwargs['conn'] = conn
            return func(self, **call_kwargs)
        finally:
            self.db_pool.putconn(conn)
    return wrapper

class DBManager:
    def __init__(self, lang):
        logger.info("DBManager initializing...")
        if any(i is None for i in ankipan_db.db_config.values()):
            raise RuntimeError('Invalid .env file')
        self.db_pool = pool.ThreadedConnectionPool(
            minconn=1,
            maxconn=N_CONNS,
            options=f"-c search_path={lang}",
            **ankipan_db.db_config,
        )
        self.lang = lang

    def get_safe_conn(self, max_attempts=N_CONNS + 1, delay=0.05):
        last = None
        backoff = delay
        for _ in range(max_attempts):
            conn = self.db_pool.getconn()
            returned = False
            try:
                if getattr(conn, "closed", 0):
                    raise psycopg2.InterfaceError("connection already closed")
                with conn.cursor() as cur:
                    cur.execute("SELECT 1")
                return conn
            except (psycopg2.OperationalError, psycopg2.InterfaceError) as e:
                last = e
                try:
                    self.db_pool.putconn(conn, close=True)
                    returned = True
                except Exception:
                    pass
                time.sleep(backoff)
                backoff = min(backoff * 2, 0.5)
                continue
            except Exception:
                try:
                    self.db_pool.putconn(conn)
                    returned = True
                except Exception:
                    pass
                raise
            finally:
                if not returned and False:
                    self.db_pool.putconn(conn)
        raise last

    @with_pool_cursor
    def get_segments_for_lemmas(self,
                                cur,
                                relative_source_paths: list[str],
                                lemmas: list[str],
                                source_category_name: str,
                                native_lang,
                                n_sentences: int = 8,
                                k_root: int = 5,
                                stride: int = 1) -> dict:
        """Return sentence-context for lemmas."""

        cur.execute("SELECT id FROM sources WHERE name=%s AND nesting_level=0", (source_category_name,))
        source_category_id_unary_list = cur.fetchone()

        if not source_category_id_unary_list:
            raise RuntimeError(f'Source Category "{source_category_name}" not defined in db')
        source_category_id = source_category_id_unary_list[0]

        lemmas = list(set(lemmas))
        path_ids: list[int] = []
        for p in relative_source_paths:
            parts = [source_category_name] + [i for i in p.strip("/").split("/") if i]
            cur.execute(f"SELECT {self.lang}.source_id_from_path(VARIADIC %s)", (parts,))
            path_ids.append(cur.fetchone()[0])

        sql = f"""
        WITH roots AS (
            SELECT * FROM UNNEST(%s::int[]) AS r(root_id)
        ),
        target_sources AS (
            SELECT sd.id
            FROM   roots
            CROSS  JOIN LATERAL source_descendants(roots.root_id) sd
        ),
        raw AS (
            SELECT
                l.lemma,
                (ts.source_id = ANY(SELECT id FROM target_sources)) AS in_target,
                r.root_name,
                src.name AS source_name,
                ts.id          AS ts_id,
                w.word,
                ts.index       AS ts_index,
                ts.source_id   AS match_source_id,
                ROW_NUMBER() OVER (
                    PARTITION BY l.lemma,
                                (ts.source_id = ANY (SELECT id FROM target_sources)),
                                r.root_name
                    ORDER BY random()
                ) AS rn_root
            FROM   lemmas l
            JOIN   words  w   ON w.lemma_id = l.id
            JOIN   words_in_text_segments wits ON wits.word_id = w.id
            JOIN   text_segments ts ON ts.id = wits.text_segment_id
            JOIN   source_root_lookup r ON r.id = ts.source_id
            JOIN   sources src ON src.id = ts.source_id
            WHERE  l.lemma = ANY(%s) AND src.id = ANY(SELECT id FROM source_descendants(%s))
        ),
        raw_limited AS (               -- cap per root for diversity
            SELECT * FROM raw
            WHERE  rn_root <= %s
        ),
        counts AS (                    -- rows per lemma inside/outside specified prioritized example sentence sources
            SELECT lemma,
                COUNT(*) FILTER (WHERE in_target)     AS inside_total,
                COUNT(*) FILTER (WHERE NOT in_target) AS outside_total
            FROM   raw_limited
            GROUP  BY lemma
        ),
        quota AS (
            SELECT lemma,
                LEAST(%s / 2, inside_total)                        AS q_in,
                LEAST(%s - LEAST(%s / 2, inside_total),
                        outside_total)                               AS q_out
            FROM   counts
        ),
        matches AS (                   -- enforce ≈50/50 quota
            SELECT *
            FROM (
                SELECT rl.*,
                    ROW_NUMBER() OVER (
                        PARTITION BY rl.lemma, rl.in_target
                        ORDER BY random()
                    ) AS rn_side
                FROM   raw_limited rl
            ) rl
            JOIN quota q USING (lemma)
            WHERE (rl.in_target AND rn_side <= q.q_in)
            OR (NOT rl.in_target AND rn_side <= q.q_out)
        )
        SELECT
            m.lemma,
            lvl1.name   AS lvl1_name,
            m.source_name,
            m.word,
            m.in_target,
            ctx.text,
            ctx.start_s,
            ctx.end_s,
            ctx.index - m.ts_index AS rel_index
        FROM matches m
        JOIN LATERAL (
                WITH RECURSIVE up AS (
                    SELECT s.id, s.parent_id, s.name, s.nesting_level
                    FROM   sources s
                    WHERE  s.id = m.match_source_id        -- start at the match
                UNION ALL
                    SELECT p.id, p.parent_id, p.name, p.nesting_level
                    FROM sources p
                    JOIN up u ON p.id = u.parent_id
                )
                SELECT name
                FROM   up
                WHERE  nesting_level = 1
                LIMIT  1
            ) lvl1 ON TRUE
        CROSS JOIN LATERAL (
            SELECT *
            FROM   text_segments ctx
            WHERE  ctx.source_id = m.match_source_id
            AND  ctx.index BETWEEN m.ts_index - %s AND m.ts_index + %s
            ORDER  BY ctx.index
        ) ctx
        ORDER  BY m.lemma, m.root_name, m.ts_id, ctx.index;
        """

        cur.execute(
            sql,(path_ids,
                lemmas,
                source_category_id,
                k_root,
                n_sentences, n_sentences, n_sentences,
                stride, stride))
        rows = cur.fetchall()

        result = {}
        for (lemma, source_root_name, source_name, word,
            in_target, text, start_s, end_s, rel_index) in rows:
            entry_type = "entries_from_known_sources" if in_target else "entries_from_unknown_sources"

            if not result.get(lemma, {}).get(entry_type, {}).get(source_root_name, {}).get(source_name):
                result.setdefault(lemma, {}).setdefault(entry_type, {}).setdefault(source_root_name, {}).setdefault(source_name, []).append({
                    "word": word,
                    'text_segments': [text],
                    'main_index': 0 if rel_index==0 else None,
                    "start_s": start_s,
                    "end_s": end_s,
                })
            else:
                result[lemma][entry_type][source_root_name][source_name][-1]['text_segments'].append(text)
                if rel_index==0:
                    result[lemma][entry_type][source_root_name][source_name][-1]['main_index'] = len(result[lemma][entry_type][source_root_name][source_name][-1]['text_segments'])-1
                result[lemma][entry_type][source_root_name][source_name][-1]['end_s'] = end_s

        # fetch cached translations (new loop because we need all text segments to fetch cached entry)
        for lemma, entries in result.items():
            for entry_type, source_data in entries.items():
                for source_root_name, sources in source_data.items():
                    for source_name, source_results in sources.items():
                        for i, source_result in enumerate(source_results):
                            cur.execute("SELECT translation from translations WHERE hash=%s AND lang = %s",
                                        (hashlib.sha256((' '.join(source_result['text_segments'])).encode("utf-8")).hexdigest(),native_lang))
                            translation_raw = cur.fetchone()
                            if translation_raw:
                                result[lemma][entry_type][source_root_name][source_name][i]['translation'] = translation_raw[0]
        return result

    def cache_translations(self, translations_by_text_segments, native_lang):
        rows = [(hashlib.sha256(o.encode("utf-8")).hexdigest(), t, native_lang) for o, t in translations_by_text_segments.items()]
        conn = self.get_safe_conn()
        with conn.cursor() as cur:
            cur.executemany(
                """
                INSERT INTO translations (hash, translation, lang)
                VALUES (%s, %s, %s)
                ON CONFLICT (hash) DO NOTHING
                """, rows)
        conn.commit()
        self.db_pool.putconn(conn)

    @with_pool_cursor
    def get_available_sources(self, cur, source_category_names: list[str] | None = None) -> dict[str, dict]:
        if source_category_names:
            cur.execute("SELECT id from sources WHERE name=%s AND nesting_level=0", (source_category_names,))
            root_ids = [id[0] for id in cur.fetchall()]
            if not root_ids:
                raise RuntimeError(
                    f'Source category name(s) not found: {source_category_names}')
        else:
            root_ids = []
        sql_roots = """
            SELECT so.id, so.name
            FROM   sources so
            WHERE parent_id IS NULL
        """
        params = []
        if root_ids:
            sql_roots += f"""
            AND so.id = ANY (
                    SELECT sd.id
                    FROM   UNNEST(%s::int[]) AS r(root_id)
                    CROSS  JOIN LATERAL source_descendants(r.root_id) sd
            )
            """
            params.append(root_ids)
        cur.execute(sql_roots, tuple(params))

        result: dict[str, dict] = {}
        root_id_by_name: dict[int, str] = {}
        for root_id, root_name in cur.fetchall():
            result[root_name] = {}
            root_id_by_name[root_id] = root_name
        if not root_id_by_name:
            return result

        cur.execute(
            """
            SELECT parent_id, name
            FROM   sources
            WHERE  parent_id = ANY(%s)
            """,
            (list(root_id_by_name.keys()),)
        )
        for parent_id, child_name in cur.fetchall():
            root_name = root_id_by_name[parent_id]
            result[root_name].setdefault(child_name, {})
        return result

    @with_pool_cursor
    def get_source_list(self, cur, source_path: str):
        path_parts = source_path.split('/')
        cur.execute(f"SELECT {self.lang}.source_id_from_path(VARIADIC %s)", (path_parts,))
        root_id = cur.fetchone()[0]
        cur.execute("SELECT metadata, lemma_counts FROM sources WHERE id = %s", (root_id,))
        root_meta, lemma_counts = cur.fetchone()
        if lemma_counts and len(lemma_counts) > 10_000:
            lemma_counts = heapq.nlargest(10_000, lemma_counts.items(), key=itemgetter(1))
        cur.execute(
            """
            SELECT name
            FROM   sources
            WHERE  parent_id = %s
            ORDER  BY name
            """, (root_id,))
        children = []
        for name in cur.fetchall():
            children.append(name[0])
        return root_meta, lemma_counts, children

    def get_source_tree_for_id(self, cur, root_id: int) -> dict[int, tuple[int | None, str]]:
        """
        Returns {id: (parent_id, name, is_leaf)} for all nodes under `root_id`
        (including the root itself).
        """
        cur.execute(
            """
            WITH RECURSIVE tree AS (
                SELECT id, parent_id, name, is_leaf
                FROM sources
                WHERE id = %s
                UNION ALL
                SELECT s.id, s.parent_id, s.name, s.is_leaf
                FROM sources s
                JOIN tree t ON s.parent_id = t.id
            )
            SELECT id, parent_id, name, is_leaf
            FROM tree;
            """,
            (root_id,),
        )
        return {row_id: (parent_id, name, is_leaf) for row_id, parent_id, name, is_leaf in cur.fetchall()}

    @with_pool_cursor
    def get_lemma_percentiles(self, cur, source_path, lemmas):
        cur.execute(f"SELECT {self.lang}.source_id_from_path(VARIADIC %s)", (source_path.split('/'),))
        root_id = cur.fetchone()[0]
        cur.execute("SELECT lemma_counts FROM sources WHERE id = %s", (root_id,))
        lemma_counts = cur.fetchone()[0]
        counter = Counter(lemma_counts)
        common = [word for word, count in counter.most_common()]
        return {lemma: (common.index(lemma) / len(counter) if counter[lemma] != 1 else 1.0)
                 for lemma in lemmas if lemma in common}

    @with_conditional_pool_conn
    def add_table_entries(
        self,
        table_name: str,
        entries: list[dict],
        *,
        ignore_duplicates: bool = False,
        return_ids: bool = True,
        conn=None):

        with conn.cursor() as cur:
            if not entries:
                return [] if return_ids else None
            cols = list(entries[0])
            values_tpl = f'({",".join(["%s"] * len(cols))})'
            values_sql = ", ".join(
                cur.mogrify(values_tpl, tuple(e[c] for c in cols)).decode()
                for e in entries)
            query = sql.SQL("""
                INSERT INTO {tbl} ({cols})
                VALUES {vals}
            """).format(
                tbl  = sql.Identifier(table_name),
                cols = sql.SQL(", ").join(map(sql.Identifier, cols)),
                vals = sql.SQL(values_sql))

            if ignore_duplicates:
                query += sql.SQL(" ON CONFLICT DO NOTHING")
            if return_ids:
                query += sql.SQL(" RETURNING id")
            cur.execute(query)
            if return_ids:
                return [row[0] for row in cur.fetchall()]

    def delete_source(self, cur, root_id: int) -> dict[str, list[int] | None]:
        sql = """
            DELETE FROM sources
            WHERE  id = %s
        """
        cur.execute(sql, (root_id,))

    @with_conditional_pool_conn
    def make_schema(self, conn=None):
        schema = self.lang

        commands = [
            f"CREATE SCHEMA IF NOT EXISTS {schema};",
            "CREATE EXTENSION IF NOT EXISTS pgcrypto;",

            f"""
            CREATE TABLE IF NOT EXISTS {schema}.translations (
                hash bytea PRIMARY KEY,
                translation TEXT NOT NULL,
                lang VARCHAR(2) NOT NULL
            );
            """,
            f"""
            CREATE TABLE IF NOT EXISTS {schema}.lemmas (
                id SERIAL PRIMARY KEY,
                lemma TEXT NOT NULL UNIQUE
            );
            """,
            f"""
            CREATE TABLE IF NOT EXISTS {schema}.words (
                id SERIAL PRIMARY KEY,
                word TEXT NOT NULL,
                pos VARCHAR(20),
                xpos VARCHAR(20),
                lemma_id INTEGER NOT NULL REFERENCES {schema}.lemmas(id),
                UNIQUE(word, pos)
            );
            """,
            f"""
            CREATE TABLE IF NOT EXISTS {schema}.sources (
                id SERIAL PRIMARY KEY,
                parent_id INT REFERENCES {schema}.sources(id) ON DELETE CASCADE,
                name TEXT NOT NULL,
                metadata jsonb,
                lemma_counts jsonb,
                nesting_level INT NOT NULL,
                is_leaf BOOL NOT NULL,
                UNIQUE (parent_id, name)
            );
            """,
            f"""
            CREATE TABLE IF NOT EXISTS {schema}.text_segments (
                id SERIAL PRIMARY KEY,
                "index" INTEGER NOT NULL,
                text TEXT NOT NULL,
                start_s INTEGER,
                end_s INTEGER,
                source_id INTEGER NOT NULL REFERENCES {schema}.sources(id) ON DELETE CASCADE
            );
            """,
            f"""
            CREATE TABLE IF NOT EXISTS {schema}.words_in_text_segments (
                word_id INTEGER NOT NULL REFERENCES {schema}.words(id),
                text_segment_id INTEGER NOT NULL REFERENCES {schema}.text_segments(id) ON DELETE CASCADE
            );
            """,

            # UNIQUE indexes needed for ON CONFLICT
            # (sources has UNIQUE(parent_id, name), lemmas has UNIQUE(lemma), words has UNIQUE(word,pos) already)
            f'CREATE UNIQUE INDEX IF NOT EXISTS uq_text_segments_source_index '
            f'  ON {schema}.text_segments (source_id, "index");',

            f'CREATE UNIQUE INDEX IF NOT EXISTS uq_wits '
            f'  ON {schema}.words_in_text_segments (word_id, text_segment_id);',

            f"""
            CREATE OR REPLACE FUNCTION {schema}.source_descendants(root_id int)
            RETURNS TABLE(id int)
            LANGUAGE sql
            SET search_path = {schema}
            AS $$
                WITH RECURSIVE d(id) AS (
                    SELECT id FROM {schema}.sources WHERE id = $1
                    UNION ALL
                    SELECT s.id FROM {schema}.sources s JOIN d ON s.parent_id = d.id
                )
                SELECT id FROM d;
            $$;
            """,
            f"""
            CREATE OR REPLACE FUNCTION {schema}.source_id_from_path(VARIADIC p_names text[])
            RETURNS int
            LANGUAGE plpgsql
            SET search_path = {schema}
            AS $$
            DECLARE
                part text;
                pid  int := NULL;
            BEGIN
                FOREACH part IN ARRAY p_names LOOP
                    SELECT id INTO pid
                    FROM   {schema}.sources
                    WHERE  parent_id IS NOT DISTINCT FROM pid
                    AND    name = part
                    LIMIT 1;
                    IF pid IS NULL THEN
                        RAISE EXCEPTION 'Path element "%" not found', part;
                    END IF;
                END LOOP;
                RETURN pid;
            END;
            $$;
            """,
            f"""
            CREATE OR REPLACE FUNCTION {schema}.lemmas_with_counts(root_id int)
            RETURNS TABLE(lemma text, cnt int)
            LANGUAGE sql
            SET search_path = {schema}
            AS $$
                SELECT l.lemma,
                    COUNT(*) AS cnt
                FROM   {schema}.source_descendants(root_id) d
                JOIN   {schema}.text_segments ts   ON ts.source_id = d.id
                JOIN   {schema}.words_in_text_segments wits ON wits.text_segment_id = ts.id
                JOIN   {schema}.words w    ON w.id = wits.word_id
                JOIN   {schema}.lemmas l   ON l.id = w.lemma_id
                GROUP  BY l.lemma
                ORDER  BY cnt DESC;
            $$;
            """,
            f"""
            CREATE OR REPLACE VIEW {schema}.source_root_lookup AS
            WITH RECURSIVE link(id, root_id, root_name) AS (
                SELECT id, id AS root_id, name AS root_name
                FROM   {schema}.sources
                WHERE  parent_id IS NULL
                UNION ALL
                SELECT s.id, l.root_id, l.root_name
                FROM   {schema}.sources s
                JOIN   link l ON s.parent_id = l.id
            )
            SELECT * FROM link;
            """,

            f"CREATE INDEX IF NOT EXISTS idx_lemmas_lemma ON {schema}.lemmas (lemma);",
            f"CREATE INDEX IF NOT EXISTS idx_words_lemma_id ON {schema}.words (lemma_id);",
            f"CREATE INDEX IF NOT EXISTS idx_text_segments_source_id ON {schema}.text_segments (source_id);",
        ]

        with conn.cursor() as cur:
            for cmd in commands:
                cur.execute(cmd)
        conn.commit()
